/*
UDP Commmunication to NTP Server
Using Day4 Example and Time NTP library example adapted for WLAN
*/
//
// Put in SSID and PWD for your LAN
//
#define SSID        "XXXXXXX"                    // Enter your SSID here
#define PASSWORD    "YYYYYYYYYYYYYYYYYYYYY"      // Enter your WLAN Password here
#define NTP_Server  "\"195.145.119.188\""        // Telekom NTP Server

#define DEBUG true

#include <Time.h> 
#include <SoftwareSerial.h>
#include <LiquidCrystal.h>

// IPAddress timeServer(132, 163, 4, 101);           // time-a.timefreq.bldrdoc.gov
// IPAddress timeServer(132, 163, 4, 102);           // time-b.timefreq.bldrdoc.gov
// IPAddress timeServer(132, 163, 4, 103);           // time-c.timefreq.bldrdoc.gov

const int timeZone = 1;                              // Central European Time
//const int timeZone = -5;                           // Eastern Standard Time (USA)
//const int timeZone = -4;                           // Eastern Daylight Time (USA)
//const int timeZone = -8;                           // Pacific Standard Time (USA)
//const int timeZone = -7;                           // Pacific Daylight Time (USA)

unsigned int localPort       = 8888;                 // local port to listen for UDP packets
unsigned int NTPlPort        = 123;                  // NTP port to receive UDP packets
const int    NTP_PACKET_SIZE = 48;                   // NTP time is in the first 48 bytes of message payload
unsigned int NTP_Data_Count  = 0;                    // Number of bytes received from NTP Server
byte         packetBuffer[NTP_PACKET_SIZE + 60];     // buffer to hold incoming & outgoing packets

time_t prevDisplay     = 0;                          // when the digital clock was displayed
time_t prevDisplay_1   = 0;                          // temp variable for display

SoftwareSerial esp8266(11, 12); // RX, TX
LiquidCrystal lcd(7, 8, 9, 10, 5, 6 );
#define       LED_WLAN 13


void setup() {
// set up the LCD's number of columns and rows:
  lcd.begin(16, 2);
// Print a message to the LCD.
  lcd.print("NTP Time Server");
//  
  delay(2000);
  Serial.begin(19200);
  Serial.println("NTP Time Server Example");
//
  digitalWrite(LED_WLAN, LOW);
//
  esp8266.begin(19200);
  
  esp8266.setTimeout(5000);
  if (sendCom("AT+RST", "ready")) {
    debug("ESP8266 reset OK");
  }

  if (configClient()) {
    debug("WLAN Client Connected");
  }
  if (configUDP()) {
    debug("NTP Timer Server Connection established");
  }
  debug("waiting for sync");
  setSyncProvider(getNtpTime);
  digitalClockDisplay();  
}

void loop() {
//
// reading / writing to the ESP8266 left in to test AT commands
// not needed for NTP functionality
//
  if (esp8266.available())
    Serial.print(char(esp8266.read()));
  if (Serial.available())
    esp8266.write(Serial.read()); 
//
  if (timeStatus() != timeNotSet) 
  {
    if (now() != prevDisplay) { //update the display only if time has changed
      prevDisplay = now();
      digitalClockDisplay();  
      Show_Time();
    }
  }
  if (minute() == 59 )
  {  
    debug("waiting for next sync");
    setSyncProvider(getNtpTime);
    debug( "Seconds since 1900 : " + String( now()));    
   }  
// 
}

//-----------------------------------------Config ESP8266------------------------------------

boolean configClient()
{
  boolean succes = true;
//
  succes &= (sendCom("AT+CWMODE=1", "OK"));
//  
  esp8266.setTimeout(20000);
  succes &= (sendCom("AT+CWJAP=\"" + String(SSID) + "\",\"" + String(PASSWORD) + "\"", "OK"));
  esp8266.setTimeout(1000);
//
  return succes;
}

boolean configUDP()
{
  boolean succes = true;
//
  succes &= (sendCom("AT+CIPMODE=0", "OK"));
  succes &= (sendCom("AT+CIPMUX=0", "OK"));
//
// this is for testing with local Packet Sender
//
//  succes &= sendCom("AT+CIPSTART=\"UDP\",\"192.168.2.118\",55005,8888,0", "OK"); // Local Packet Server
//
// using Telekom NTP server
//
  succes &= sendCom("AT+CIPSTART=\"UDP\"," + String(NTP_Server) + ",123,8888,0", "OK");   // NTP Server

//  
  return succes;
}

//-----------------------------------------------NTP code---------------------------------------------------------

time_t getNtpTime()
{

  unsigned long secsSince1900;
  byte          offset = 40;
  
  debug("Transmit NTP Request");
  sendUTPpacket();
//
  if ( NTP_Data_Count >= ( NTP_PACKET_SIZE - 1) )
  {  
// convert four bytes starting at location 40 to a long integer
    secsSince1900 =  (unsigned long)packetBuffer[offset + 0] << 24;
    secsSince1900 |= (unsigned long)packetBuffer[offset + 1] << 16;
    secsSince1900 |= (unsigned long)packetBuffer[offset + 2] << 8;
    secsSince1900 |= (unsigned long)packetBuffer[offset + 3];
    digitalWrite(LED_WLAN, HIGH);
    return secsSince1900 - 2208988800UL + timeZone * SECS_PER_HOUR;   // UNIX Date
//   return secsSince1900 + timeZone * SECS_PER_HOUR;                    // NTP  Date
  }
  else
  {
    Serial.println("No NTP Response ");
    digitalWrite(LED_WLAN, LOW);
    return 0;                                                 // return 0 if unable to get the time
  }
}



// prepare an NTP request to the time server and send to the given

void sendUTPpacket()
{
// set all bytes in the buffer to 0
  memset(packetBuffer, 0, NTP_PACKET_SIZE + 2);
// Initialize values needed to form NTP request
// (see URL above for details on the packets)
  packetBuffer[0]   = 0b11100011;   // LI, Version, Mode
  packetBuffer[1]   = 0;            // Stratum, or type of clock
  packetBuffer[2]   = 6;            // Polling Interval
  packetBuffer[3]   = 0xEC;         // Peer Clock Precision
// 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer[12]  = 49;
  packetBuffer[13]  = 0x4E;
  packetBuffer[14]  = 49;
  packetBuffer[15]  = 52;
//
// if we need a cr/lf
//
  packetBuffer[NTP_PACKET_SIZE    ]  = 0x0D;
  packetBuffer[NTP_PACKET_SIZE + 1]  = 0x0A;
//  
// all NTP fields have been given values, now
// you can send a packet requesting a timestamp:  
//  
  if ( send_BIN( packetBuffer, NTP_PACKET_SIZE ))
    debug("NTP ready");  
}

//-----------------------------------------------Controll ESP-----------------------------------------------------

boolean sendCom(String command, char respond[])
{
  esp8266.println(command);
  if (esp8266.findUntil(respond, "ERROR"))
  {
    return true;
  }
  else
  {
    debug("ESP SEND ERROR: " + command);
    return false;
  }
}

String sendCom(String command)
{
  esp8266.println(command);
  return esp8266.readString();
}

//---------------------------------------------Send Binary Data to ESP-----------------------------------------------------

boolean send_BIN(byte Msg[], int MsgSize)

{
  boolean succes = true;

  succes &= sendCom("AT+CIPSEND=" + String(MsgSize + 0), ">");    
  if (succes)
  {
    succes &= send_BINCom(Msg, MsgSize);
  }
  return succes;
}

boolean send_BINCom(byte command[], int commandSize)

{
int  index =0;
// send command, convert byte to character  
//
  for ( index = 0; index < commandSize; index++ )
     esp8266.print(char(command[index]));
// 
// check for success
//
  index = 0;
  if (esp8266.find( "+IPD,48:"))
//
// data net payload will be 48 bytes ( always ?? )
// this can be done better by reading the integer 
// and write the number of data bytes received
//
  {
    debug("ESP binary sent OK and data received");
    uint32_t beginWait = millis();
    while (millis() - beginWait < 500) 
    {
      if ( esp8266.available() >= commandSize )
      {
        for ( index = 0; index < commandSize; index++ )
        {  packetBuffer[index] = byte(esp8266.read());
        }  
        NTP_Data_Count = index; 
        if (esp8266.find( "/r/n" )) {};   // this will clear the input buffer
        return true;
      }
    }
    debug("NTP Server Timeout");
    return false;
  }
  else
  {
    debug("ESP binary send ERROR");
    return false;
  }

//
}
//-------------------------------------------------Debug Functions------------------------------------------------------

void debug(String Msg)
{
  if (DEBUG)
  {
    Serial.println(Msg);
  }
}

/*-------- Time Display Code ---------*/

void digitalClockDisplay(){
  // digital clock display of the time
  Serial.print(hour());
  printDigits(minute());
  printDigits(second());
  Serial.print(" ");
  Serial.print(day());
  Serial.print(".");
  Serial.print(month());
  Serial.print(".");
  Serial.print(year()); 
  Serial.println(); 
}

void printDigits(int digits){
  // utility for digital clock display: prints preceding colon and leading 0
  Serial.print(":");
  if(digits < 10)
    Serial.print('0');
  Serial.print(digits);
}

void printBuffer( int index )
{
//
// for test purposes, print packer buffer ( send & receive )
//
  Serial.print(" Packet buffer:");
  
  for ( int i = 0; i < index ; i++ )
  {
    Serial.print(char(packetBuffer[i]));
    Serial.print(" ");
  }  
  Serial.println();
}

void Show_Time()
{
  String shour, sminute, ssecond;
  lcd.setCursor(0, 1);
  
  // print the number of seconds since reset:
  if (hour() <= 9) shour = "0" + String(hour()); else shour = String(hour()); // adjust for 0-9
  if (minute() <= 9) sminute = "0" + String(minute());  else sminute = String(minute());
  if (second() <= 9) ssecond = "0" + String(second());  else ssecond = String(second());
  lcd.print( shour + ":" + sminute + ":" + ssecond );
}

